/*
  ==========================================
  quintCode by C.S. Allin, copyright 2025
  My website: http://www.crunchyslug.com

  quintCode and all associated files are
  released under the C.S. Allin quintCode 
  License v1.0. see license.txt for details. 
  to summarize, however, you can do whatever 
  you want with quintCode except for sell it
  (or sell products based on it) or steal credit
  for it.

  if you like quintCode and think it's worth
  money, please visit my webpage and throw
  money at me so I can make a living. thanks.

  I can confirm this sourcecode compiles under
  MS-DOS, Windows and garden variety Unix-like
  operating systems. if it doesn't compile under
  your platform of choice then you'll probably
  need to tinker with the header files.

  this interpreter automatically converts
  forward slashes to backslashes in file paths
  and vice versa. if you're using a platform
  that doesn't use slashes in its file paths
  at all and this feature will cause problems
  for you, go to the function makeTarg() and
  remove the slash flipping code.
  ==========================================
*/

#if defined(__APPLE__) || defined(__unix__) || defined(__linux__)
#define UNIXLIKE_OS
#endif

#ifdef UNIXLIKE_OS
#include <sys/stat.h>
#include <sys/types.h>
/*
  if the quintCode interpreter is compiled on a Unix-like OS,
  these permissions will be applied to the webpages it generates.
  change them as you please.
*/
#define WEBPAGE_PERMS S_IRUSR | S_IWUSR | S_IRGRP |S_IROTH
#endif

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h> /*time.h and calls to srand() and rand() are only used in
		    this sourcecode for the purposes of randomizing elements
		    in the demo webpages. you can remove them.*/

#define quintCodeName(codeString) strstr(quintCode, codeString)
#define insertFile(fpath) _insertFile(fpath, tmp)
#define insertString(str) _insertString(str, tmp)

char buffer[150]; /*when a path is read from qsources.txt it goes in this buffer - make it bigger if you need to!*/
char quintCode[6] = {'\0','\0','\0','\0','\0','\0'};

struct targ {
	char *path;
	struct targ *next;
};

struct targ *targList = NULL;

char *sourcelist = "qsources.txt";
char *tempfile = "qtemp.txt";

void _insertFile(char *path, FILE *tmp);
void _insertString(char *str, FILE *tmp);

/*-----------------------
  -----------------------
  START HERE! handleQuintCode() is the only function you really
  need to bother with when programming the quintCode interpreter
  to do your bidding. you should only alter the other functions
  if you want to drastically change the way the interpreter works.
  or if you want to add more rigorous error checking.
  -----------------------
  -----------------------
*/
char handleQuintCode(FILE *tmp){
	int random = 0;
	random = rand() % 5;

	/*printf("Valid quintCode: %s\n", quintCode);*/

	if(quintCodeName("BGCOL")){
		if(random == 0){
			insertString("bgcolor=blue");
		}else if(random == 1){
			insertString("bgcolor=green");
		}else if(random == 2){
			insertString("bgcolor=red");
		}else if(random == 3){
			insertString("bgcolor=orange");
		}else if(random == 4){
			insertString("bgcolor=purple");
		}
		
		return 1; /*ALWAYS return 1 after you handle a quintCode!*/
	}

	if(quintCodeName("GREET")){
		random = rand()%4;

		if(random == 0){
			insertFile("greet1.txt");
		}else if(random == 1){
			insertFile("greet2.txt");
		}else if(random == 2){
			insertFile("greet3.txt");
		}else{
			insertFile("greet4.txt");
		}
		
		return 1;
	}
	
	return 0;
}

/*-----------------------
  -----------------------
  writes a file to qtemp.txt
*/
void _insertFile(char *path, FILE *tmp){
	char b = 0;
	FILE *f = NULL;

	f = fopen(path, "rb");
	if(f == NULL){
		printf("\ninsertFile can't open %s.\n", path);
		exit(0);
	}

	while(fread(&b, 1, 1, f)){
		fwrite(&b, 1, 1, tmp);
	}

	fclose(f);
}

/*-----------------------
  -----------------------
  writes a string to qtemp.txt
*/
void _insertString(char *string, FILE *tmp){
	unsigned int len = strlen(string)-1, i = 0;
	while(i <= len){
		fwrite(string+i, 1, 1, tmp);
		i++;
	}
}

/*-----------------------
  -----------------------*/
void zeroEntireBuffer(){
	unsigned char i = 0;

	while(i <= 149){
		buffer[i] = '\0';
		i++;
	}
}

/*-----------------------
  -----------------------*/
void resetQuintCode(){
	char i = 0;
	while(quintCode[i] != '\0'){
		quintCode[i] = '\0';
	}
}

/*-----------------------
  -----------------------*/
void resetBuffer(){
	unsigned char i = 0;

	while(buffer[i] != '\0'){
		buffer[i] = '\0';
		i++;
	}
}

/*-----------------------
  -----------------------*/
void makeTarg(){
	struct targ *t = NULL;
	unsigned int len = 0, i = 0;

	t = calloc(1, sizeof(struct targ));
	t->path = calloc(1, strlen(buffer)+1);

	strcpy(t->path, buffer);

	/*
	  SLASH FLIPPING CODE
	  unix uses / in its paths while DOS/WINDOWS uses \
	  this code flips / to \ as needed. remove this code
	  if it's going to cause any issues for you.
	*/
	len = strlen(t->path) -1;
	while(i < len){
		#ifdef UNIXLIKE_OS
		if(t->path[i] == '\\'){
			t->path[i] = '/';
		}
		#endif

		#ifndef UNIXLIKE_OS
		if(t->path[i] == '/'){
			t->path[i] = '\\';
		}
		#endif

		i++;
	}

	t->next = targList;
	targList = t;
}

/*-----------------------
  -----------------------
  loads paths from qsources.txt and builds
  a linked list of out of them.
*/
void buildTargList(){
	char b = 0;
	FILE *f = NULL;
	unsigned int i = 0, bindex=0;

	f = fopen(sourcelist, "r");

	while(fread(&b, 1, 1, f)){

		/*
		  DOS doesn't care if \r is in the path but
		  unix/linux balks at it. so filter out \r.
		  and also filter \n while we're at it.
		*/
		if(b != '\r' && b != '\n'){
			buffer[i] = b;
			bindex++;
		}

		if(b == '\n'){
			makeTarg();
			resetBuffer();
			i = 0;
			continue;
		}

		i++;
	}

	if(buffer[0] != '\0'){
		makeTarg();
	}
}

/*-----------------------
  -----------------------
  checks to see if something that looks like it
  might be a quintCode actually is a quintCode
*/
char validateQuintCode(FILE *fsrc, FILE *ftmp){
	char len = 0, b = 0, result = 0;
	long offset = 0;
	resetQuintCode();

	while(len <= 4 && fread(&b, 1, 1, fsrc) == 1){
		fwrite(&b, 1, 1, ftmp);
		/*1 to 9, uppercase A to Z*/
		if((b >= 48 && b <= 57) || (b >= 65 && b <= 90)){
			quintCode[len] = b;
			len++;
		}else{
			break;
		}
	}

	if(len < 4){
		resetQuintCode();
		return 0;
	}

	/*
	  what character is immediately after the quintCode?
	*/
	result = fread(&b, 1, 1, fsrc);
	fwrite(&b, 1, 1, ftmp);

	/*
	  not a valid quintCode if it's under 5 characters long,
	  isn't followed by - or a space, or is at the end of a file.
	*/
	if((b != '-' && b != ' ') || result == 0) {
		resetQuintCode();
		return 0;
	}

	/*
	  the quintCode appears to be in an HTML comment
	  like <!--;THISS--> so seek past the "-->"
	  at the end.
	*/

	if(b == '-'){
		fread(&b, 1, 1, fsrc);
		fwrite(&b, 1, 1, ftmp);
		fread(&b, 1, 1, fsrc);
		fwrite(&b, 1, 1, ftmp);
	}

	/*
	  if there's a newline after the quintCode then
	  the user probably wants it to stay there. seek
	  past it.
	*/

	offset = ftell(fsrc);
	fread(&b, 1, 1, fsrc);

	if(b == '\n'){
		/*fread(&b, 1, 1, fsrc);*/
		fwrite(&b, 1, 1, ftmp);
		return 1;
	}

	fseek(fsrc, offset, SEEK_SET);

	return 1;
}

/*-----------------------
  -----------------------
  this is used for finding the string "END"
  in the source file. not really suitable
  for anything else.
*/
char findStringInFile(char *str, FILE *f){
	unsigned int len = strlen(str), i = 0, offset=0, hits = 0;
	char b = 0;

	offset = ftell(f);

	while(str[i] != '\0'){
		if(fread(&b, 1, 1, f) < 1){
			break;
		}

		if(str[i] == b){
			hits++;
		}
		i++;
	}

	if(hits == len){
		fseek(f, offset, SEEK_SET);
		return 1;
	}

	fseek(f, offset, SEEK_SET);
	return 0;
}

/*-----------------------
  -----------------------
  iterate through all the quintCodes in the source file,
  call handleQuintCode(), take care of writing to qtemp.txt
  and reading from source files, etc.
*/
void processQuintCodes(char *path){
	char b = 0;
	char *end = "END\0";
	long offset = 0;
	FILE *fsource = NULL, *ftemp = NULL;
	unsigned int quintCodeCount = 0, endStatementCount = 0;

	fsource = fopen(path, "rb");

	if(fsource == NULL){
		printf("\nprocessQuintCodes can't open %s.\n", path);
		exit(0);
	}

	ftemp = fopen(tempfile, "wb");

	if(ftemp == NULL){
		exit(0);
	}

	while(fread(&b, 1, 1, fsource)){
		fwrite(&b, 1, 1, ftemp);

		if(b == ';'){
			/*
			remember where we found ; in the source file
			because we might need to know this later.
			this is what "offset" is for.
			*/
			/*offset = ftell(fsource);*/

			/*
			validateQuintCode will seek to just after
			a valid quintCode. this is where the user's
			specified file starts being written to ftemp.
			*/
			if(validateQuintCode(fsource, ftemp)){

				/*
				user writes dynamic data to ftemp based on
				whatever the quintCode is
				*/
				if(handleQuintCode(ftemp) == 0){
					continue;
				}

				quintCodeCount++;

				/*
				find the ;END statement that marks where the
				dynamic zone ends.
				*/

				while(fread(&b, 1, 1, fsource)){

					if(b == ';'){
						/*
						we found ; which means this might be the
						;END statement. let's find out. remember
						where the character just after ; is by saving
						the file position in "offset".
						*/
						offset = ftell(fsource);

						if(findStringInFile(end, fsource) == 1){

							/*
							we've definitely encountered the
							string ;END in the source file.
							is it inside an HTML comment?
							rewind to the character right before
							; and see what it is. if it's -
							then assume ;END is inside an HTML
							comment like so: <!--;END-->
							*/

							fseek(fsource, offset -2, SEEK_SET);

							fread(&b, 1, 1, fsource);
							if(b == '-'){
								/*
								seek to the beginning of the END
								comment. this is where fsource
								will resume being written to
								ftemp.
								*/
								endStatementCount++;
								fseek(fsource, offset -6, SEEK_SET);
								fread(&b, 1, 1, fsource);
								if(b == '\n' || b == ' '){
									fseek(fsource, offset -6, SEEK_SET);
								}

								break;
							}else{
								/*
								;END statement isn't in an HTML comment.
								seek to ;. fsource resumes being written...
								well, you know the drill
								*/
								endStatementCount++;
								fseek(fsource, offset-2, SEEK_SET);
								break;
							}
						}
					}
				}
			}
		}
	}


	if(quintCodeCount != endStatementCount){
		printf("Mismatch between number of quintCodes and number of ;END statements\nin source file %s. Quitting.\n\n", path);
		exit(0);
	}

	/*
	  qtemp.txt is now the updated version of the page. overwrite the
	  old version with this updated version.
	*/

	fclose(fsource);
	fclose(ftemp);

	remove(path);

	#ifdef UNIXLIKE_OS
	chmod(tempfile, WEBPAGE_PERMS);
	#endif

	rename(tempfile, path);
}

/*-----------------------
  -----------------------*/
int main(int argc, char *argv[]){
	struct targ *t = NULL;

	srand(time(NULL));

	if(argc == 2){
		sourcelist = argv[1];
	}else if(argc == 3){
		sourcelist = argv[1];
		tempfile = argv[2];
	}

	zeroEntireBuffer();
	buildTargList();

	t = targList;

	while(t != NULL){
		processQuintCodes(t->path);
		t = t->next;
	}

	printf("\nquintCode intepreter: All pages generated.\n");

	return 0;
}